#include <QApplication>
#include <QComboBox>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QShortcut>
#include <QSpinBox>
#include <QVBoxLayout>

#include <QResizeEvent>
#include <QScreen>
#include <QWindow>

#include <QDebug>
#include <QTimer>

#include <algorithm>

bool optStretch = true;

struct dumpScreen {
    explicit dumpScreen(const QScreen *s) : m_screen(s) {}

    const QScreen *m_screen;
};

QDebug operator<<(QDebug d, const dumpScreen &ds)
{
    QDebugStateSaver s(d);
    d.noquote();
    d.nospace();
    d << ds.m_screen->name() << "\" " << ds.m_screen->geometry()
        << ", dpr=" << ds.m_screen->devicePixelRatio()
        << ' ' << ds.m_screen->logicalDotsPerInch() << "dpi";
    return d;
}

struct dumpScreens
{
    explicit dumpScreens(const QWindow *referenceWindow = nullptr) : m_referenceWindow(referenceWindow) {}

    const QWindow *m_referenceWindow;
};

QDebug operator<<(QDebug d, const dumpScreens &ds)
{
    QDebugStateSaver s(d);
    d.noquote();
    d.nospace();
    const QScreen *refScreen = ds.m_referenceWindow ? ds.m_referenceWindow->screen() : nullptr;
    for (auto s : QGuiApplication::screens())
        d << (s == refScreen ? " *  " : "    ") << dumpScreen(s) << '\n';
    return d;
}

class EventFilter : public QObject {
public:
    EventFilter(QObject *p) : QObject(p) {}

    bool eventFilter(QObject *, QEvent *);

private:
    static void reportEvent(const QObject *o, const QEvent *);
};

bool EventFilter::eventFilter(QObject *o, QEvent *e)
{
    const QEvent::Type t = e->type();
    switch (t) {
    case QEvent::Resize:
        reportEvent(o, e);
        break;
    default:
        break;
    }
    return false;
}

void EventFilter::reportEvent(const QObject *o, const QEvent *e)
{
    static int n = 0;
    if (o->isWindowType()) {
        const auto *w = static_cast<const QWindow *>(o);
        qDebug().noquote().nospace() << '#' << n++ << ' ' << w->objectName() << " dpr="
            << w->devicePixelRatio() << ' ' << e;
    } else if (o->isWidgetType()) {
        const auto *w = static_cast<const QWidget *>(o);
        if (w->isWindow()) {
            qDebug().noquote().nospace() << '#' << n++ << ' ' << w->objectName() << ' '
                << w->logicalDpiX() << "DPI, dpr=" << w->devicePixelRatioF() << ' ' << e;
        }
    }
}

static QString title()
{
    return QLatin1String(QT_VERSION_STR) + QLatin1Char(' ')
            + QGuiApplication::platformName();
}

class Widget : public QWidget {
    Q_OBJECT
public:
    Widget();

public slots:
    void dump()
    {
        qDebug().noquote() << __FUNCTION__
           << logicalDpiX() << "DPI, dpr=" << devicePixelRatioF()
           << geometry() << '\n' << dumpScreens(windowHandle()) << '\n';
    }

    void screenChanged()
    {
        qDebug().noquote() << __FUNCTION__ << geometry() << dumpScreen(windowHandle()->screen());
    }

    void appScreenAdded(QScreen *screen)
    {
        connectScreen(screen);
        qDebug().noquote() << __FUNCTION__ << '\n' << dumpScreens(windowHandle());
    }

    void appScreenRemoved()
    {
        qDebug().noquote() << __FUNCTION__ << '\n' << dumpScreens(windowHandle());
    }

    void appPrimaryScreenChanged()
    {
        qDebug().noquote() << __FUNCTION__ << '\n' << dumpScreens(windowHandle());
    }

    void screenGeometryChanged(const QRect &geometry)
    {
        auto screen = qobject_cast<const QScreen *>(sender());
        qDebug().noquote() << __FUNCTION__ << (screen ? screen->name() : sender()->objectName()) << geometry;
    }

    void screenPhysicalSizeChanged(const QSizeF &size)
    {
        auto screen = qobject_cast<const QScreen *>(sender());
        qDebug().noquote() << __FUNCTION__ << (screen ? screen->name() : sender()->objectName()) << size;
    }

    void screenPhysicalDotsPerInchChanged(qreal dpi)
    {
        auto screen = qobject_cast<const QScreen *>(sender());
        qDebug().noquote() << __FUNCTION__ << (screen ? screen->name() : sender()->objectName()) << dpi;
    }

    void screenLogicalDotsPerInchChanged(qreal dpi)
    {
        auto screen = qobject_cast<const QScreen *>(sender());
        qDebug().noquote() << __FUNCTION__ << (screen ? screen->name() : sender()->objectName()) << dpi;
    }

private:
    void connectScreen(QScreen *s);
};

Widget::Widget()
{
    setObjectName("Top");
    for (auto s : QGuiApplication::screens())
        connectScreen(s);

    auto hl = new QHBoxLayout(this);

    auto vl = new QVBoxLayout;
    vl->addWidget(new QLabel(title()));
    auto cb = new QComboBox;
    cb->addItems({"i1", "i2"});
    vl->addWidget(cb);
    auto sp = new QSpinBox;
    sp->setRange(0, 100);
    vl->addWidget(sp);
    vl->addWidget(new QLineEdit("bla"));

    auto nestedHl = new QHBoxLayout;
    auto b = new QPushButton("Dump");
    connect(b, &QAbstractButton::clicked, this, &Widget::dump);
    b->setShortcut(Qt::Key_D + Qt::CTRL);
    nestedHl->addWidget(b);
    b = new QPushButton("Quit");
    connect(b, &QAbstractButton::clicked, this, &QWidget::close);
    b->setShortcut(Qt::Key_Q + Qt::CTRL);
    nestedHl->addWidget(b);
    vl->addLayout(nestedHl);

    if (optStretch)
        vl->addStretch();
    hl->addLayout(vl);
    if (optStretch)
        hl->addStretch();
}

void Widget::connectScreen(QScreen *s)
{
    connect(s, &QScreen::geometryChanged, this, &Widget::screenGeometryChanged);
    connect(s, &QScreen::physicalSizeChanged, this, &Widget::screenPhysicalSizeChanged);
    connect(s, &QScreen::physicalDotsPerInchChanged, this, &Widget::screenPhysicalDotsPerInchChanged);
    connect(s, &QScreen::logicalDotsPerInchChanged, this, &Widget::screenLogicalDotsPerInchChanged);
}

int main(int argc, char *argv[])
{
    QStringList args;
    std::copy(argv + 1, argv + argc, std::back_inserter(args));

    if (args.contains("-n")) {
        qDebug("AA_DisableHighDpiScaling");
        QCoreApplication::setAttribute(Qt::AA_DisableHighDpiScaling);
    } else {
        qDebug("AA_EnableHighDpiScaling");
        QCoreApplication::setAttribute(Qt::AA_UseHighDpiPixmaps);
        QCoreApplication::setAttribute(Qt::AA_EnableHighDpiScaling);
    }

    QApplication a(argc, argv);
    if (!args.contains("-e"))
        a.installEventFilter(new EventFilter(&a));
    if (args.contains("-s"))
        optStretch = false;

    Widget w;
    QObject::connect(&a, &QGuiApplication::screenAdded, &w, &Widget::appScreenAdded);
    QObject::connect(&a, &QGuiApplication::screenRemoved, &w, &Widget::appScreenRemoved);
    QObject::connect(&a, &QGuiApplication::primaryScreenChanged, &w, &Widget::appPrimaryScreenChanged);

    QString t = title();

    qDebug().noquote() << t << args << '\n' << dumpScreens();

    w.setWindowTitle(t);
    w.show();
    QObject::connect(w.windowHandle(), &QWindow::screenChanged, &w, &Widget::screenChanged);

    return a.exec();
}

#include "main.moc"
